import asyncio
import itertools
import logging
import math
import random

import pylru
from asyncio_dgram.aio import DatagramStream
from asyncio_dgram.aio import Protocol

from .source import messages

MAX_SIZE_32 = 2 ** 31 - 1

logger = logging.getLogger('sqproxy.transport')


class UnknownRequest(Exception):
    pass


class BrokenPacketError(Exception):  # FIXME?
    def __init__(self, raw_data, addr):
        self.raw_data = raw_data
        self.addr = addr


class SourceDatagramStream(DatagramStream):
    FRAGMENT_MAX_SIZE = 1200
    MAX_FRAGMENTS_PER_PACKET = 100

    def __init__(self, transport, recvq, excq, drained):
        super().__init__(transport, recvq, excq, drained)
        self.fragments = pylru.lrucache(size=1024)

    def handle_fragments(self, packet):
        header = messages.Header.decode(packet)
        if header['split'] != messages.SPLIT:
            return packet

        fragment = messages.Fragment.decode(header.raw_tail)
        packet_id = fragment['message_id']

        if packet_id in self.fragments:
            fragment_count, packet_fragments = self.fragments[packet_id]
        else:
            fragment_count = fragment['fragment_count']
            if fragment_count > self.MAX_FRAGMENTS_PER_PACKET:
                logger.warning(
                    'Packet fragment count more than our limit: %s > %s.' 'Packet will be trimmed',
                    fragment_count,
                    self.MAX_FRAGMENTS_PER_PACKET,
                )
                fragment_count = self.MAX_FRAGMENTS_PER_PACKET

            packet_fragments = []
            self.fragments[packet_id] = fragment_count, packet_fragments

        fragment_ids = {f['fragment_id'] for f in packet_fragments}
        if fragment['fragment_id'] not in fragment_ids:
            packet_fragments.append(fragment)

        if len(packet_fragments) < fragment_count:
            return None

        self.fragments.pop(packet_id, None)
        packet_fragments.sort(key=lambda f: f['fragment_id'])
        return b''.join(f.raw_tail for f in packet_fragments)

    async def send_packet(self, packet, addr=None, split_size=FRAGMENT_MAX_SIZE):
        if len(packet) <= split_size:
            await self.send(packet, addr)
            return

        message_id = random.randint(1, MAX_SIZE_32)
        fragment_count = math.ceil(len(packet) / split_size)  # type: int
        mtu = split_size - (4 + (4 + 1 + 1 + 2))  # MAX_SIZE - (packet header + fragment header)

        packet_iter = iter(packet)

        for fragment_id in range(fragment_count):
            fragment_tail = bytes(itertools.islice(packet_iter, mtu))

            fragment_header = messages.Fragment().encode(
                message_id=message_id,
                fragment_count=fragment_count,
                fragment_id=fragment_id,
                mtu=mtu,
                split_header=True,
            )

            await self.send(b''.join((fragment_header, fragment_tail)), addr)

    async def recv_packet(self):
        while True:
            data, addr = await super().recv()

            try:
                data = self.handle_fragments(data)
            except messages.BrokenMessageError:
                raise BrokenPacketError(data, addr)

            if data is None:
                # data not ready
                continue

            return data, addr

    async def send_bytes(self, data, addr=None):
        """Alias for `send()`, use it instead `send()`
         to explicit difference from `send_packet()`
        """
        return await self.send(data, addr)

    async def recv_bytes(self):
        """Alias for `recv()`, use it instead `recv()`
         to explicit difference from `recv_packet()`
        """
        return await self.recv()

    @staticmethod
    def _decode(packet, msg_classes):
        header = messages.Header.decode(packet)
        packet = header.raw_tail

        for cls in msg_classes:
            msg = cls.decode(packet, default=None)
            if msg is not None:
                return msg

        return None

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        self.close()


class SourceDatagramServer(SourceDatagramStream):
    request_message_classes = (
        messages.InfoRequest,
        messages.PlayersRequest,
        messages.RulesRequest,
    )

    def decode_request(self, packet):
        return self._decode(packet, msg_classes=self.request_message_classes)

    async def recv_packet(self):
        try:
            data, addr = await super().recv_packet()
        except BrokenPacketError as exc:
            return None, exc.raw_data, exc.addr

        message = self.decode_request(data)
        return message, data, addr


class SourceDatagramClient(SourceDatagramStream):
    response_message_classes = (
        messages.InfoResponse,
        messages.PlayersResponse,
        messages.RulesResponse,
        messages.GetChallengeResponse,
    )

    def decode_response(self, packet):
        return self._decode(packet, msg_classes=self.response_message_classes)

    async def recv_packet(self):
        while True:
            try:
                data, addr = await super().recv_packet()
            except BrokenPacketError as exc:
                return None, exc.raw_data, exc.addr

            message = self.decode_response(data)
            return message, data, addr


async def bind(addr) -> SourceDatagramServer:
    """
    Bind a socket to a local address for datagrams.  The socket will be either
    AF_INET or AF_INET6 depending upon the type of address specified.

    @param addr - For AF_INET or AF_INET6, a tuple with the the host and port to
                  to bind; port may be set to 0 to get any free port.
    @return     - A SourceDatagramServer instance
    """
    loop = asyncio.get_event_loop()
    recvq = asyncio.Queue()
    excq = asyncio.Queue()
    drained = asyncio.Event()

    transport, protocol = await loop.create_datagram_endpoint(
        lambda: Protocol(recvq, excq, drained), local_addr=addr, reuse_address=False
    )

    return SourceDatagramServer(transport, recvq, excq, drained)


async def connect(addr) -> SourceDatagramClient:
    """
    Connect a socket to a remote address for datagrams.  The socket will be
    either AF_INET or AF_INET6 depending upon the type of host specified.

    @param addr - For AF_INET or AF_INET6, a tuple with the the host and port to
                  to connect to.
    @return     - A SourceDatagramClient instance
    """
    loop = asyncio.get_event_loop()
    recvq = asyncio.Queue()
    excq = asyncio.Queue()
    drained = asyncio.Event()

    transport, protocol = await loop.create_datagram_endpoint(lambda: Protocol(recvq, excq, drained), remote_addr=addr)

    return SourceDatagramClient(transport, recvq, excq, drained)



import asyncio
import logging
import os
from ipaddress import IPv4Address
from ipaddress import ip_address

import pyroute2

from . import config

logger = logging.getLogger(__name__)


def _get_addr_interface(addr: IPv4Address, ipdb=pyroute2.IPDB()):
    for idx, addresses in ipdb.ipaddr.items():
        for ifaddr, prefix in addresses:
            if ip_address(ifaddr) == addr:
                return ipdb.by_index[idx]['ifname']
    return None


def get_ebpf_program_run_args():
    args = []

    interface = None
    for server_name, server in config.servers:
        server_interface = _get_addr_interface(server.network.bind_ip)
        assert server_interface is not None, f"Can't get interface name for {server.network.bind_ip}"

        if interface is None:
            interface = server_interface

        if server_interface != interface:
            raise config.ConfigurationError(
                'Different interfaces dont supported yet: ' f'{server_interface} != {interface}'
            )

        args += ['-p', f'{server.network.server_port}:{server.network.bind_port}']

    args += ['-i', interface]

    return args


async def run_ebpf_redirection():
    program = config.ebpf.executable

    args = [config.ebpf.script_path.name]
    args += get_ebpf_program_run_args()

    logger.info('Run %s %s', program, args)

    process = await asyncio.create_subprocess_exec(
        program, *args, stdout=asyncio.subprocess.PIPE, cwd=config.ebpf.script_path.parent.as_posix(),
    )

    while True:
        data = await process.stdout.readline()
        if data:
            logger.info(data.decode().rstrip())

        if process.stdout.at_eof():
            break

    retcode = process.returncode
    if retcode != os.EX_OK:
        logger.exception('eBPF redirection exit with code %s', retcode)
        raise RuntimeError

    logger.info('eBPF redirection normally exit with 0 code')



import logging.config
import logging.handlers
import os


class BetterRotatingFileHandler(logging.handlers.RotatingFileHandler):
    def _open(self):
        os.makedirs(os.path.dirname(self.baseFilename), exist_ok=True)
        return super()._open()


def get_file_handler_opts(filename: str, level: str):
    return {
        'class': BetterRotatingFileHandler.__module__ + '.' + BetterRotatingFileHandler.__qualname__,
        'maxBytes': 1024 * 1024,
        'backupCount': 3,
        'level': level,
        'formatter': 'verbose',
        'filename': filename,
    }


def setup_logging(debug_filename: str = None, error_filename: str = None):
    if debug_filename is None:
        debug_filename = '/dev/null'

    if error_filename is None:
        error_filename = '/dev/null'

    # fmt: off
    logging.config.dictConfig(
        {
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {
                'verbose': {
                    'format': '%(asctime)s [%(levelname)s] [%(name)s] %(message)s',
                },
            },
            'handlers': {
                'null': {
                    'level': 'DEBUG',
                    'class': 'logging.NullHandler',
                },
                'console': {
                    'class': 'logging.StreamHandler',
                    'level': 'DEBUG',
                    'formatter': 'verbose',
                },
                'debug_file': get_file_handler_opts(debug_filename, 'DEBUG'),
                'error_file': get_file_handler_opts(error_filename, 'ERROR'),
            },
            'loggers': {
                '': {'handlers': ['console', 'debug_file', 'error_file'], 'level': 'DEBUG', 'propagate': False},
                'PidFile': {'handlers': ['null'], 'propagate': False},
            },
        }
    )
    # fmt: on



import click


@click.group()
def sqproxy():
    """Basic entrypoint"""


@sqproxy.command()
def run():
    """Run SQProxy process"""
    from .__main__ import run

    run()


if __name__ == '__main__':
    sqproxy()



# Copyright (C) 2013 Oliver Ainsworth

import collections
import functools
import socket
import struct

from . import util

NO_SPLIT = -1
SPLIT = -2

_missing = object()


class BrokenMessageError(Exception):
    pass


class BufferExhaustedError(BrokenMessageError):
    def __init__(self, message='Incomplete message'):
        BrokenMessageError.__init__(self, message)


def on_broken_default(func):
    @functools.wraps(func)
    def wrap(*args, **kw):
        default = kw.pop('default', _missing)

        try:
            return func(*args, **kw)
        except BrokenMessageError:
            if default is not _missing:
                return default
            raise

    return wrap


def on_header_required(func):
    @functools.wraps(func)
    def wrap(*args, **kw):
        split_header = kw.pop('split_header', False)
        nosplit_header = not split_header

        result = func(*args, **kw)

        split = (nosplit_header and NO_SPLIT) or (split_header and SPLIT) or _missing

        if split is _missing:
            raise RuntimeError

        header = Header().encode(split=split)
        result = b''.join((header, result))

        return result

    return wrap


def use_self_default(func):
    def use_default(self, value=None, values=None):
        if value is None:
            return func(self, self.default_value, values)
        return func(self, value, values)

    return use_default


def needs_buffer(func):
    def needs_buffer(self, buffer, *args, **kwargs):
        if len(buffer) == 0:
            raise BufferExhaustedError
        return func(self, buffer, *args, **kwargs)

    return needs_buffer


def int2ip(value):
    return socket.inet_ntoa(struct.pack('i', value))


class MessageField(object):

    fmt = None
    validators = []

    def __init__(self, name, optional=False, default_value=None, validators=None):
        """
            name -- used when decoding messages to set the key in the
                returned dictionary

            optional -- whether or not a field value must be provided
                when encoding

            default_value -- if optional is False, the value that is
                used if none is specified

            validators -- list of callables that return False if the
                value they're passed is invalid
        """

        if validators is None:
            validators = []

        if self.fmt is not None:
            if self.fmt[0] not in '@=<>!':
                self.format = '<' + self.fmt
            else:
                self.format = self.fmt

        self.name = name
        self.optional = optional
        self._value = default_value
        self.validators = self.__class__.validators + validators

    @property
    def default_value(self):
        if self.optional:
            if self._value is not None:
                return self._value
        raise ValueError("Field '{fname}' is not optional".format(fname=self.name))

    def validate(self, value):
        for validator in self.validators:
            try:
                if not validator(value):
                    raise ValueError
            except Exception:
                raise BrokenMessageError("Invalid value ({}) for field '{}'".format(value, self.name))
        return value

    @use_self_default
    def encode(self, value, values=None):
        try:
            return struct.pack(self.format, self.validate(value))
        except struct.error as exc:
            raise BrokenMessageError(exc)

    @needs_buffer
    def decode(self, buffer, values=None):
        """
            Accepts a string of raw bytes which it will attempt to
            decode into some Python object which is returned. All
            remaining data left in the buffer is also returned which
            may be an empty string.

            Also acecpts a second argument which is a dictionary of the
            fields that have been decoded so far (i.e. occurs before
            this field in `fields` tuple). This allows the decoder to
            adapt it's funtionality based on the value of other fields
            if needs be.

            For example, in the case of A2S_PLAYER resposnes, the field
            `player_count` needs to be accessed at decode-time to determine
            how many player entries to attempt to decode.
        """

        field_size = struct.calcsize(self.format)
        if len(buffer) < field_size:
            raise BufferExhaustedError

        field_data = buffer[:field_size]
        left_overs = buffer[field_size:]

        try:
            return (self.validate(struct.unpack(self.format, field_data)[0]), left_overs)
        except struct.error as exc:
            raise BrokenMessageError(exc)


class ByteField(MessageField):
    fmt = 'B'


class StringField(MessageField):
    fmt = 's'

    @use_self_default
    def encode(self, value, values=None):
        return value.encode('utf8') + b'\x00'

    @needs_buffer
    def decode(self, buffer, values=None):
        terminator = buffer.find(b'\x00')
        if terminator == -1:
            raise BufferExhaustedError('No string terminator')
        field_size = terminator + 1
        field_data = buffer[: field_size - 1]
        left_overs = buffer[field_size:]
        return self.validate(field_data.decode('utf8', 'ignore')), left_overs


class ShortFieldLE(MessageField):  # little-endian
    fmt = '<h'


class ShortFieldBE(MessageField):  # big-endian
    fmt = '>h'


class LongFieldLE(MessageField):  # little-endian
    fmt = '<l'


class LongFieldBE(MessageField):  # big-endian
    fmt = '>l'


class FloatField(MessageField):
    fmt = 'f'


class PlatformField(ByteField):
    @needs_buffer
    def decode(self, buffer, values=None):
        byte, remnant_buffer = super(PlatformField, self).decode(buffer, values)
        return util.Platform(byte), remnant_buffer


class ServerTypeField(ByteField):
    @needs_buffer
    def decode(self, buffer, values=None):
        byte, remnant_buffer = super(ServerTypeField, self).decode(buffer, values)
        return util.ServerType(byte), remnant_buffer


class IpAddrField(LongFieldBE):
    validators = [lambda x: int2ip(x)]


class MessageArrayField(MessageField):
    """
        Represents a nested message within another message that is
        repeated a given number of time (often defined within the
        same message.)
    """

    def __init__(self, name, element, count=None):
        """
            element -- the Message subclass that will attempt to be decoded

            count -- ideally a callable that returns the number of
                'elements' to attempt to decode; count must also present
                a 'minimum' attribute which is minimum number of elements
                that must be decoded or else raise BrokenMessageError

                If count isn't callable (e.g. a number) it will be
                wrapped in a function with the minimum attribute set
                equal to the given 'count' value

                Helper static methods all(), value_of() and at_least()
                are provided which are intended to be used as the
                'count' argument, e.g.

                MessageArrayField("", SubMessage, MessageArrayField.all())

                ... will decode all SubMessages within the buffer
        """

        MessageField.__init__(self, name)
        if count is None:
            count = self.all()
        # Coerces the count argument to be a callable. For example,
        # in most cases count would be a Message.value_of(), however
        # if an integer is provided it will be wrapped in a lambda.
        self.count = count
        if not hasattr(count, '__call__'):

            def const_count(values=None):
                return count

            const_count.minimum = count
            self.count = const_count
        self.element = element

    def encode(self, elements, values=None):
        if values is None:
            values = {}

        buf = []
        for i, element in enumerate(elements):
            if not isinstance(element, self.element):
                raise BrokenMessageError(
                    'Element {} ({}) is not instance of {}'.format(i, element, self.element.__name__)
                )
            if i + 1 > self.count(values):
                raise BrokenMessageError('Too many elements')
            buf.append(element.encode())
        if len(buf) < self.count.minimum:
            raise BrokenMessageError('Too few elements')
        return b''.join(buf)

    def decode(self, buffer, values=None):
        if values is None:
            values = {}

        entries = []
        count = 0
        while count < self.count(values):
            # Set start_buffer to the beginning of the buffer so that in
            # the case of buffer exhaustion it can return from the
            # start of the entry, not half-way through it.
            #
            # For example if you had the fields:
            #
            #       ComplexField =
            #           LongField
            #           ShortField
            #
            #       MessageArrayField(ComplexField,
            #                         count=MessageArrayField.all())
            #       ByteField()
            #
            # When attempting to decode the end of the buffer FF FF FF FF 00
            # the first four bytes will be consumed by LongField,
            # however ShortField will fail with BufferExhaustedError as
            # there's only one byte left. However, there is enough left
            # for the trailing ByteField. So when ComplexField
            # propagates ShortField's BufferExhaustedError the buffer will
            # only have the 00 byte remaining. The exception if caught
            # and buffer reverted to FF FF FF FF 00. This is passed
            # to ByteField which consumes one byte and the reamining
            # FF FF FF 00 bytes and stored as message payload.
            #
            # This is very much an edge case. :/
            start_buffer = buffer
            try:
                entry = self.element.decode(buffer)
                buffer = entry.raw_tail
                entries.append(entry)
                count += 1
            except (BufferExhaustedError, BrokenMessageError) as exc:
                # Allow for returning 'at least something' if end of
                # buffer is reached.
                if count < self.count.minimum:
                    raise BrokenMessageError(exc)
                buffer = start_buffer
                break
        return entries, buffer

    @staticmethod
    def value_of(name):
        """
            Reference another field's value as the argument 'count'.
        """

        def field(values, f=None):
            f.minimum = values[name]
            return values[name]

        field.__defaults__ = (field,)
        return field

    @staticmethod
    def all():
        """
            Decode as much as possible from the buffer.

            Note that if a full element field cannot be decoded it will
            return all entries decoded up to that point, and reset the
            buffer to the start of the entry which raised the
            BufferExhaustedError. So it is possible to have addtional
            fields follow a MessageArrayField and have
            count=MessageArrayField.all() as long as the size of the
            trailing fields < size of the MessageArrayField element.
        """

        i = [1]

        def all_(values=None):
            i[0] = i[0] + 1
            return i[0]

        all_.minimum = -1
        return all_

    @staticmethod
    def at_least(minimum):
        """
            Decode at least 'minimum' number of entries.
        """

        i = [1]

        def at_least(values=None):
            i[0] = i[0] + 1
            return i[0]

        at_least.minimum = minimum
        return at_least


class MessageDictField(MessageArrayField):
    """
        Decodes a series of key-value pairs from a message. Functionally
        identical to MessageArrayField except the results are returned as
        a dictionary instead of a list.
    """

    def __init__(self, name, key_field, value_field, count=None):
        """
            key_field and value_field are the respective components
            of the name-value pair that are to be decoded. The fields
            should have unique name strings. Tt is assumed that the
            key-field comes first, followed by the value.

            count is the same as MessageArrayField.
        """

        element = type('KeyValueField', (Message,), {'fields': (key_field, value_field)})
        self.key_field = key_field
        self.value_field = value_field
        MessageArrayField.__init__(self, name, element, count)

    def decode(self, buffer, values=None):
        entries, buffer = MessageArrayField.decode(self, buffer, values)
        entries_dict = {}
        for entry in entries:
            entries_dict[entry[self.key_field.name]] = entry[self.value_field.name]
        return entries_dict, buffer


class Message(collections.abc.MutableMapping):

    fields = ()

    def __init__(self, raw_tail=None, **values):
        self.raw_tail = raw_tail
        self.values = values

    def __getitem__(self, key):
        return self.values[key]

    def __setitem__(self, key, value):
        self.values[key] = value

    def __delitem__(self, key):
        del self.values[key]

    def __len__(self):
        return len(self.values)

    def __iter__(self):
        return iter(self.values)

    def encode(self, **field_values):
        values = dict(self.values, **field_values)
        buf = []
        for field in self.fields:
            buf.append(field.encode(values.get(field.name, None), values))
        return b''.join(buf)

    @classmethod
    @on_broken_default
    def decode(cls, packet):
        buffer = packet
        values = {}
        for field in cls.fields:
            values[field.name], buffer = field.decode(buffer, values)
        return cls(buffer, **values)


class Header(Message):

    fields = (LongFieldLE('split', validators=[lambda x: x in [SPLIT, NO_SPLIT]]),)


class Packet(Message):
    """Message with Header"""

    encode = on_header_required(Message.encode)


class Fragment(Packet):

    fields = (
        LongFieldLE('message_id'),
        ByteField('fragment_count'),
        ByteField('fragment_id'),  # 0-indexed
        ShortFieldLE('mtu'),
    )

    @property
    def is_compressed(self):
        # check MSB (Most Significant Bit)
        return bool(self['message_id'] & (1 << 16))


# TODO: FragmentCompressionData


class InfoRequest(Packet):

    fields = (
        ByteField('request_type', True, 0x54, validators=[lambda x: x == 0x54]),
        StringField('payload', True, 'Source Engine Query'),
    )


class InfoResponse(Packet):

    fields = (
        ByteField('response_type', validators=[lambda x: x == 0x49]),
        ByteField('protocol'),
        StringField('server_name'),
        StringField('map'),
        StringField('folder'),
        StringField('game'),
        ShortFieldLE('app_id'),
        ByteField('player_count'),
        ByteField('max_players'),
        ByteField('bot_count'),
        ServerTypeField('server_type'),
        PlatformField('platform'),
        ByteField('password_protected'),  # BooleanField
        ByteField('vac_enabled'),  # BooleanField
        StringField('version')
        # TODO: EDF
    )


class GetChallengeResponse(Packet):

    fields = (ByteField('response_type', True, 0x41, validators=[lambda x: x == 0x41]), LongFieldLE('challenge'))


class PlayersRequest(Packet):

    fields = (ByteField('request_type', True, 0x55, validators=[lambda x: x == 0x55]), LongFieldLE('challenge'))


class PlayerEntry(Packet):

    fields = (ByteField('index'), StringField('name'), LongFieldLE('score'), FloatField('duration'))


class PlayersResponse(Packet):

    fields = (
        ByteField('response_type', validators=[lambda x: x == 0x44]),
        ByteField('player_count'),
        MessageArrayField('players', PlayerEntry, MessageArrayField.value_of('player_count')),
    )


class RulesRequest(Packet):

    fields = (ByteField('request_type', True, 0x56, validators=[lambda x: x == 0x56]), LongFieldLE('challenge'))


class RulesResponse(Packet):

    fields = (
        # A2S_RESPONSE misteriously seems to add a FF FF FF FF
        # long to the beginning of the response which isn't
        # mentioned on the wiki.
        #
        # Behaviour witnessed with TF2 server 94.23.226.200:2045
        # LongFieldLE("long"),
        ByteField('response_type', validators=[lambda x: x == 0x45]),
        ShortFieldLE('rule_count'),
        MessageDictField('rules', StringField('key'), StringField('value'), MessageArrayField.value_of('rule_count')),
    )


# For Master Server
class MSAddressEntryPortField(MessageField):
    fmt = '!H'


class MSAddressEntryIPField(MessageField):
    @needs_buffer
    def decode(self, buffer, values=None):
        if len(buffer) < 4:
            raise BufferExhaustedError
        field_data = buffer[:4]
        left_overs = buffer[4:]
        return '.'.join(str(b) for b in struct.unpack(b'<BBBB', field_data)), left_overs


class MasterServerRequest(Packet):

    fields = (ByteField('request_type', True, 0x31), ByteField('region'), StringField('address'), StringField('filter'))


class MSAddressEntry(Message):

    fields = (MSAddressEntryIPField('host'), MSAddressEntryPortField('port'))

    @property
    def is_null(self):
        return self['host'] == '0.0.0.0' and self['port'] == 0


class MasterServerResponse(Packet):

    fields = (
        # The first two fields are always FF FF FF FF and 66 0A
        # and can be ignored.
        MSAddressEntryIPField('start_host'),
        MSAddressEntryPortField('start_port'),
        MessageArrayField('addresses', MSAddressEntry, MessageArrayField.all()),
    )



import logging
import pathlib
import typing
from ipaddress import IPv4Address

import sentry_sdk
import yaml
from pydantic import AnyHttpUrl
from pydantic import BaseModel
from pydantic import BaseSettings
from pydantic import Extra
from sentry_sdk.integrations.logging import LoggingIntegration

from . import __version__
from .dict_merge import dict_merge
from .logging import setup_logging

__all__ = (
    'NetworkModel',
    'ServerModel',
    'Settings',
    'ConfigurationError',
)

logger = logging.getLogger('sqproxy.config')


class ConfigurationError(Exception):
    pass


class NetworkModel(BaseModel):
    server_ip: IPv4Address
    server_port: int
    bind_ip: IPv4Address
    bind_port: int

    class Config:
        extra = Extra.forbid


class ServerModel(BaseModel):
    network: NetworkModel
    a2s_info_cache_lifetime: int = 5
    a2s_rules_cache_lifetime: int = 5
    a2s_players_cache_lifetime: int = 1
    src_query_port_lifetime: int = 10
    no_a2s_rules: bool = False

    class Config:
        extra = Extra.forbid


class EBPFModel(BaseModel):
    enabled: bool = False
    executable: str = 'python2'
    script_path: pathlib.Path = './src-ebpf/redirect.py'

    class Config:
        extra = Extra.forbid


class Settings(BaseSettings):
    sentry_dsn: typing.Optional[AnyHttpUrl] = None
    confdir_0: pathlib.Path = '/etc/sqproxy/conf.d/'
    confdir_1: pathlib.Path = './conf.d/'
    error_log: pathlib.Path = '/var/log/sqproxy/error.log'
    debug_log: pathlib.Path = '/var/log/sqproxy/debug.log'
    debug_log_enabled: bool = False
    piddir: typing.Optional[pathlib.Path] = None

    class Config:
        env_file = '.env'
        env_prefix = 'SQPROXY_'

    def get_merged_config_data(self):
        return load_configs(iter_config_files(self.confdir_0, self.confdir_1))


def _apply_defaults(target, defaults):
    target.update(dict_merge(defaults, target))


def _get_config(data: dict, global_defaults) -> typing.Tuple[typing.Dict, typing.Dict]:
    """Get full and ready to use config
    """
    defaults = data.pop('defaults', None) or {}

    servers_data = data.get('servers')
    if not servers_data:
        return data, defaults

    for server in servers_data.values():
        for g_defaults in global_defaults:
            _apply_defaults(server, g_defaults['values'])

        if defaults and defaults.get('values'):
            _apply_defaults(server, defaults['values'])

    return data, defaults


def iter_config_files(*confdirs):
    for confdir in confdirs:
        if not confdir.exists():
            logger.info('Confdir not found: %s', confdir.absolute().as_posix())
            continue

        if not confdir.is_dir():
            logger.warning('Confdir expected to be a directory, not a file: %s', confdir.absolute().as_posix())
            continue

        for file in confdir.iterdir():
            if file.is_file() and file.name.endswith('.yaml'):
                logger.info('Found config: %s', file.as_posix())
                yield file
            else:
                logger.debug('Found non-config file, ignore: %s', file.as_posix())


def load_configs(paths: typing.Iterable[pathlib.Path]):
    configs = []
    global_defaults = []

    ebpf_configured = False

    for path in paths:
        with path.open() as fp:
            config, config_defaults = _get_config(yaml.full_load(fp), global_defaults)

            if 'ebpf' in config:
                if ebpf_configured:
                    raise ConfigurationError('eBPF already configured')
                ebpf_configured = True

            if config_defaults.get('global', False):
                global_defaults.append(config_defaults)

            if config:
                configs.append(config)

    whole_config = {}
    for config in configs:
        whole_config.update(config)

    return whole_config


def setup(settings_: Settings = None):
    global settings
    if settings_ is None:
        logger.info('Re-read settings')
        settings_ = Settings()

    settings = settings_

    if settings.sentry_dsn:
        sentry_logging = LoggingIntegration(
            level=logging.DEBUG,  # Capture info and above as breadcrumbs
            event_level=logging.ERROR,  # Send errors as events
        )
        sentry_sdk.init(
            dsn=settings.sentry_dsn, integrations=[sentry_logging], release=__version__,
        )

    debug_filename = settings.debug_log_enabled and settings.debug_log.as_posix() or None

    setup_logging(
        debug_filename=debug_filename, error_filename=settings.error_log.as_posix(),
    )

    if settings.sentry_dsn:
        logger.debug('Sentry enabled')
    else:
        logger.debug('Sentry disabled')

    merged_config_data = settings.get_merged_config_data()

    servers_ = merged_config_data.get('servers')
    if servers_:
        global servers
        servers = [(name, ServerModel.parse_obj(server)) for name, server in servers_.items()]

    ebpf_ = merged_config_data.get('ebpf')
    if ebpf_:
        global ebpf
        ebpf = EBPFModel.parse_obj(ebpf_)


settings = Settings()
servers = None  # type: typing.Optional[typing.List[typing.Tuple[str, ServerModel]]]
ebpf = None  # type: typing.Optional[EBPFModel]


setup(settings)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['source_query_proxy', 'source_query_proxy.source']

package_data = \
{'': ['*']}

install_requires = \
['async-timeout>=3.0,<4.0',
 'asyncio_dgram>=1.0,<2.0',
 'backoff>=1.10,<2.0',
 'click>=7.0,<8.0',
 'pid>=2.2,<4.0',
 'pydantic[dotenv]>=1.4,<2.0',
 'pylru>=1.2.0,<2.0.0',
 'pyroute2>=0.5.10,<0.6.0',
 'python-dotenv>=0.10.3,<0.13.0',
 'pyyaml>=5.2,<6.0',
 'sentry-sdk>=0.14.3,<0.15.0',
 'uvloop==0.14.0']

entry_points = \
{'console_scripts': ['sqproxy = source_query_proxy.cli:sqproxy']}

setup_kwargs = {
    'name': 'source-query-proxy',
    'version': '1.1.0',
    'description': 'Async proxy for Source Engine Query Protocol',
    'long_description': '\nsource-query-proxy\n==================\n\nCredits\n-------\n\nSource Engine messages inspired by **Python-valve**\nhttps://github.com/serverstf/python-valve\n\nPrerequisites\n-------------\n\nPython 3.7 or above\n\nYou can use `pyenv <https://github.com/pyenv/pyenv>`_ to install any version of Python without root privileges\n\nInstalling\n----------\n\n.. code-block:: bash\n\n    pip install source-query-proxy==1.1.0\n\nConfiguring\n-----------\n\nsqproxy search configs in ``/etc/sqproxy/conf.d`` and ``./conf.d`` directories.\nYou should place your config files only in this directories.\n\nFor more info see `examples <examples/conf.d>`_\n\nRun\n---\n\n.. code-block:: bash\n\n    sqproxy run\n\n\nRun with eBPF\n-------------\n\nhttps://github.com/spumer/source-query-proxy-kernel-module/tree/master/src-ebpf\n\n1. Download eBPF script and copy ``src-ebpf`` folder to target working directory\n\n2. Install requirements https://github.com/spumer/source-query-proxy-kernel-module/tree/master/src-ebpf/README.md\n\n3. Enable eBPF in config (see examples/00-globals.yaml)\n\n4. Run\n\n.. code-block:: bash\n\n    sqproxy run\n\n\nDevelopment\n-----------\n\n.. code-block:: bash\n\n    git clone https://github.com/spumer/source-query-proxy.git\n    cd source-query-proxy\n    poetry install\n',
    'author': 'spumer',
    'author_email': 'spumer-tm@yandex.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/spumer/source-query-proxy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
https://gist.github.com/CMeza99/5eae3af0776bef32f945f34428669437
"""


def dict_merge(base_dct, merge_dct):
    """
    Recursive dict merge.
    Args:
        base_dct (dict) onto which the merge is executed
        merge_dct (dict): base_dct merged into base_dct
        add_keys (bool): whether to add new keys
    Returns:
        dict: updated dict
    """
    rtn_dct = base_dct.copy()
    rtn_dct.update(
        {
            key: dict_merge(rtn_dct[key], merge_dct[key])
            if isinstance(rtn_dct.get(key), dict) and isinstance(merge_dct[key], dict)
            else merge_dct[key]
            for key in merge_dct.keys()
        }
    )
    return rtn_dct
